import apache_beam as beam
import apache_beam.io.fileio as beam_io
from pathlib import Path
from apache_beam.testing.test_pipeline import TestPipeline as BeamTestPipeline
from apache_beam.testing.util import assert_that, equal_to

from klay_beam.transforms import LoadWithLibrosa


def test_LoadWithLibrosa():
    this_file = Path(__file__)
    data_dir = this_file.parent / "test_data" / "audio"
    wav_pattern = str(data_dir / "*.wav")
    ogg_pattern = str(data_dir / "*.ogg")
    mp3_pattern = str(data_dir / "*.mp3")
    patterns = [wav_pattern, ogg_pattern, mp3_pattern]

    # This list of filenames is a hard-coded copy of the list of files in the
    # tests/test_data/audio directory. These files were generated by the
    # bin/create_test_audio_files.py script. Each file is 0.5 seconds at
    # 44.1kHz.
    data = {
        "test_mono.mp3": (22050,),  # Annoyingly, this is what librosa returns for mono
        "test_quad.ogg": (4, 22050),
        "test_quad.wav": (4, 22050),
        "test_stereo.mp3": (2, 22050),
        "test_stereo.ogg": (2, 22050),
        "test_stereo_16bit.wav": (2, 22050),
        "test_stereo_24bit.wav": (2, 22050),
        "test_stereo_32bit.wav": (2, 22050),
        "test_stereo_64bit.wav": (2, 22050),
    }

    def verify_data(element):
        filename, audio, sr = element
        basename = Path(filename).name
        expected_shape = data[basename]

        assert sr == 44100
        assert audio.shape == expected_shape

    expected_filenames = [str(Path(data_dir / datum)) for datum in data.keys()]

    with BeamTestPipeline() as p:
        # Apply the custom transform
        readable_files = p | beam.Create(patterns) | beam.io.fileio.MatchAll()

        # Extract file names from the metadata for assertion
        file_names = readable_files | "Extract File Names" >> beam.Map(
            lambda metadata: metadata.path
        )

        # Check if the readable_files match the expected file names
        assert_that(file_names, equal_to(expected_filenames))

        # Load the audio files using librosa
        (
            readable_files
            | beam_io.ReadMatches()
            | beam.ParDo(LoadWithLibrosa(target_sr=None, mono=False))
            | beam.Map(verify_data)
        )


def test_fail_LoadWithLibrosa():
    this_file = Path(__file__)
    data_dir = this_file.parent / "test_data" / "audio" / "corrupt"
    wav_pattern = str(data_dir / "*.wav")
    ogg_pattern = str(data_dir / "*.ogg")
    mp3_pattern = str(data_dir / "*.mp3")
    patterns = [wav_pattern, ogg_pattern, mp3_pattern]

    # This list of filenames is a hard-coded copy of the list of files in the
    # tests/test_data/audio directory. These files were generated by the
    # bin/create_test_audio_files.py script. Each file is 0.5 seconds at
    # 44.1kHz.
    data = {"test_corrupt.mp3": ()}

    expected_filenames = [str(Path(data_dir / datum)) for datum in data.keys()]

    with BeamTestPipeline() as p:
        # Apply the custom transform
        readable_files = p | beam.Create(patterns) | beam.io.fileio.MatchAll()

        # Extract file names from the metadata for assertion
        file_names = readable_files | "Extract File Names" >> beam.Map(
            lambda metadata: metadata.path
        )

        # Check if the readable_files match the expected file names
        assert_that(file_names, equal_to(expected_filenames))

        # Load the audio files using librosa
        audio, failed = (
            readable_files
            | beam_io.ReadMatches()
            | beam.ParDo(LoadWithLibrosa(target_sr=None, mono=False)).with_outputs(
                "failed", main="audio"
            )
        )

        def verify_failed(element):
            assert element == 1

        (
            failed
            | "Count Failed" >> beam.combiners.Count.Globally()
            | "Check Failed" >> beam.Map(verify_failed)
        )

        def verify_audio_failed(element):
            assert element == 0

        (
            audio
            | "Count Audio" >> beam.combiners.Count.Globally()
            | "Check Audio Failed" >> beam.Map(verify_audio_failed)
        )



import os.path
from pathlib import PurePosixPath
import re


def move(filename, source_dir, target_dir):
    """Move a file from source_dir to target_dir, preserving the relative
    directory structure. Should work across local paths and GCS URIs. Throw if
    input is not in source_dir.

    ```
    # Example usage:
    move(
        'gs://klay-datasets/audios/00/1009600.mp3',
        'gs://klay-datasets/audios/',
        '/somewhere/else/'
    ) == '/somewhere/else/00/1009600.mp3'
    ```
    """
    relative_source_filename = PurePosixPath(filename).relative_to(source_dir)
    return os.path.join(target_dir, relative_source_filename)

    # If we wanted to avoid os.path.join, we could detect URIs and handle them
    # separately. However, this is more complicated and also uses the _replace
    # method (which is technically private). See:
    # https://stackoverflow.com/questions/38552253/change-urlparse-path-of-a-url

    # from urllib.parse import urlparse, urlunparse
    # result = str(PurePosixPath(target_dir) / relative_source_filename)
    # target_dir_is_uri = urlparse(target_dir).scheme != ""
    # if target_dir_is_uri:
    #     target_uri = urlparse(target_dir)
    #     new_path = str(
    #         PurePosixPath(target_uri.path) / PurePosixPath(relative_source_filename)
    #     )
    #     target_uri = target_uri._replace(path=new_path)
    #     result = urlunparse(target_uri)

    # return result


assert os.path.sep == "/", "os.path.join (in get_target_path) breaks on Windows"


def remove_suffix(path: str, suffix: str):
    if path.endswith(suffix):
        return path[: -len(suffix)]
    return path


def remove_suffix_pattern(path: str, suffix_pattern: str):
    regex = re.compile(r"{}".format(suffix_pattern))
    match = regex.search(path)
    if match is None:
        return path
    else:
        return remove_suffix(path, match[0])



from typing import Optional, Union, Tuple, List
import logging
import numpy as np
import apache_beam as beam
import apache_beam.io.fileio as beam_io
import pathlib
import io
import scipy

from .torch_utils import TORCH_AVAILABLE, ensure_torch_available


if TORCH_AVAILABLE:
    import torch
    import torchaudio


class LoadWithTorchaudio(beam.DoFn):
    """Use torchaudio to load audio files to tensors

    NOTES:

    - torchaudio depends on libavcodec, which can be installed with:
    `conda install 'ffmpeg<5'`. See:
    https://github.com/pytorch/audio/issues/2363#issuecomment-1179089175


    - Torchaudio supports loading in-memory (file-like) files since at least
    v0.9.0. See: https://pytorch.org/audio/0.9.0/backend.html#load


    Note that generally, custom functions have a few requirements that help them
    work well in on distributed runners. They are:
        - The function should be thread-compatible
        - The function should be serializable
        - Recommended: the function be idempotent

    For details about these requirements, see the Apache Beam documentation:
    https://beam.apache.org/documentation/programming-guide/#requirements-for-writing-user-code-for-beam-transforms
    """

    def setup(self):
        # This will be executed only once when the pipeline starts. This is
        # where you would create a lock or queue for global resources.
        ensure_torch_available()
        pass

    def process(self, readable_file: beam_io.ReadableFile):  # type: ignore
        """
        Given an Apache Beam ReadableFile, return a `(input_filename, a, sr)` tuple where
            - `input_filename` is a string
            - `a` is a pytorch Tensor
            - `sr` is an int

        For a stereo audio file named '/path/to.some/file.key.mp3', return
        ```
        (
            '/path/to.some/file.key.mp3',
            tensor([[0.1, 0.2, 0.3], [0.1, 0.2, 0.3]]),
            44100
        )
        ```
        """
        path = pathlib.Path(readable_file.metadata.path)

        # get the file extension without a period in a safe way
        ext_without_dot = path.suffix.lstrip(".")
        ext_without_dot = None if ext_without_dot == "" else ext_without_dot

        audio_tensor: torch.Tensor
        sr: int

        # try loading the audio file with torchaudio, but catch RuntimeError,
        # which are thrown when torchaudio can't load the file.
        logging.info("Loading: {}".format(path))
        try:
            with readable_file.open(mime_type="application/octet-stream") as file_like:
                audio_tensor, sr = torchaudio.load(file_like, format=ext_without_dot)
        except (RuntimeError, OSError) as e:
            # We don't want to log the stacktrace, but for debugging, here's how
            # we could access it:
            #
            # import traceback
            # tb_str = traceback.format_exception(
            #     etype=type(e), value=e, tb=e.__traceback__
            # )
            logging.warning(f"Error loading audio: {path}")
            return [beam.pvalue.TaggedOutput("failed", (str(path), e))]

        C, T = audio_tensor.shape
        duration_seconds = T / sr
        logging.info(f"Loaded {duration_seconds:.3f} second {C}-channel audio: {path}")

        return [
            (readable_file.metadata.path, audio_tensor, sr),
            beam.pvalue.TaggedOutput("duration_seconds", duration_seconds),
        ]


class ResampleTorchaudioTensor(beam.DoFn):
    """Resample an audio to a new sample rate. Accepts a `(key, a, sr)` tuple
    were:

    - `key` is a string
    - `a` is a 2D torch.Tensor or numpy.ndarray with audio in the last dimension
    - `sr` is an int

    The return value will also be a `(key, a, sr)` tuple, but 'a' will always be
    torch.Tensor instance.
    """

    MAX_AUDIO_CHANNELS = 128

    def __init__(
        self,
        target_sr: int,
        source_sr_hint: Optional[int] = None,
        output_numpy: bool = False,
    ):
        assert isinstance(
            target_sr, int
        ), f"target_sr must be an int (found {target_sr})"
        self.target_sr = target_sr
        self.source_sr_hint = source_sr_hint
        self.output_numpy = output_numpy
        self.resample = None

    def setup(self):
        ensure_torch_available()
        if self.source_sr_hint is not None:
            self.resample = torchaudio.transforms.Resample(
                self.source_sr_hint, self.target_sr
            )

    def process(  # type: ignore
        self,
        audio_tuple: Tuple[str, Union["torch.Tensor", np.ndarray], int],
    ):
        key, audio, source_sr = audio_tuple

        # check if audio_tensor is a numpy array or a torch tensor
        if isinstance(audio, np.ndarray):
            audio = torch.from_numpy(audio)

        channels, _ = audio.shape
        if channels > self.MAX_AUDIO_CHANNELS:
            raise ValueError(
                f"audio_tensor ({key}) must have {self.MAX_AUDIO_CHANNELS} or "
                f"fewer channels (found {channels})"
            )

        resampled_audio: torch.Tensor

        if source_sr == self.target_sr:
            resampled_audio = audio
            logging.info(
                f"Skipping resample because source was already {self.target_sr}: {key}"
            )
        elif (
            self.resample is not None
            and source_sr == self.source_sr_hint  # type: ignore[unreachable]
        ):
            resampled_audio = self.resample(audio)  # type: ignore[unreachable]
            logging.info(
                f"Resampled {source_sr} to {self.target_sr} (cached method): {key}"
            )
        else:
            resample = torchaudio.transforms.Resample(source_sr, self.target_sr)
            resampled_audio = resample(audio)
            logging.info(
                f"Resampled {source_sr} to {self.target_sr} (uncached method): {key}"
            )

        if self.output_numpy:
            resampled_audio = resampled_audio.numpy()

        return [(key, resampled_audio, self.target_sr)]


def convert_audio(wav: "torch.Tensor", sr: int, target_sr: int, target_channels: int):
    """Copied from encodec"""
    ensure_torch_available()
    if wav.ndim == 1:
        wav.unsqueeze_(0)
    assert wav.shape[0] in [1, 2], "Audio must be mono or stereo."
    if target_channels == 1:
        wav = wav.mean(0, keepdim=True)
    elif target_channels == 2:
        *shape, _, length = wav.shape
        wav = wav.expand(*shape, target_channels, length)
    elif wav.shape[0] == 1:
        wav = wav.expand(target_channels, -1)
    wav = torchaudio.transforms.Resample(sr, target_sr)(wav)
    return wav


def tensor_to_bytes(
    audio_tuple: Tuple[str, Union["torch.Tensor", np.ndarray], int]
) -> List[Tuple[str, bytes, int]]:
    fname, audio, sr = audio_tuple
    if isinstance(audio, torch.Tensor):
        audio = audio.detach().cpu().numpy()

    buf = io.BytesIO()
    scipy.io.wavfile.write(buf, sr, audio)
    buf.seek(0)
    wav_data = buf.read()

    return [(fname, wav_data, sr)]



import pytest
from klay_beam.path import move, remove_suffix_pattern


def test_move_preserves_directory_structure():
    assert (
        move(
            "gs://audio/00/1009600.mp3",
            "gs://audio/",
            "/somewhere/else/",
        )
        == "/somewhere/else/00/1009600.mp3"
    ), "gs:// to local"

    assert (
        move(
            "/audio/00/1009600.mp3",
            "/audio/",
            "/somewhere/else/",
        )
        == "/somewhere/else/00/1009600.mp3"
    ), "local to local"

    assert (
        move(
            "/a//audio/00/1009600.mp3",
            "/a/audio//",
            "/somewhere/else/",
        )
        == "/somewhere/else/00/1009600.mp3"
    ), "local to local with extra slashes"

    assert (
        move(
            "/audio/00/1009600.mp3",
            "/audio/",
            "gs://somewhere/else/",
        )
        == "gs://somewhere/else/00/1009600.mp3"
    ), "local to gs://"


def test_move_fails():
    with pytest.raises(ValueError):
        move(
            "gs://audio/00/1009600.mp3",
            "gs://audio/01",
            "/somewhere/else/",
        )

    with pytest.raises(ValueError):
        move(
            "gs://audio/00/1009600.mp3",
            "//audio/00",
            "/somewhere/else/",
        )


@pytest.mark.parametrize(
    "path, pattern, exp",
    [
        ("path/to/audio.drums.wav", r".drums(-\d)?\.wav", "path/to/audio"),
        ("path/to/audio.drums.wav", ".drums.wav", "path/to/audio"),
        ("/path/to/audio.drums.wav", r".drums(-\d)?\.wav", "/path/to/audio"),
        ("gs://path/to/audio.drums.wav", r".drums(-\d)?\.wav", "gs://path/to/audio"),
        ("path/to/audio.drums-1.wav", r".drums(-\d)?\.wav", "path/to/audio"),
    ],
)
def test_remove_suffix_pattern(path, pattern, exp):
    got = remove_suffix_pattern(path, pattern)
    assert got == exp



from typing import Tuple, Optional
import logging


def torch_check() -> Tuple[bool, Optional[ImportError]]:
    torch_available, torch_import_error = False, None
    try:
        import torch
        import torchaudio

        if torch.__version__ < "1.0.0":
            raise ImportError(
                "Incompatible version of torch is installed. Install version 1.0.0 or newer."
            )
        if torchaudio.__version__ < "0.8.0":
            raise ImportError(
                "Incompatible version of torchaudio is installed. Install version 0.8.0 or newer."
            )

        torch_available = True
    except ImportError as e:
        torch_import_error = e
        logging.info(f"torch is not available: {e}")
    return torch_available, torch_import_error


TORCH_AVAILABLE, TORCH_IMPORT_ERROR = torch_check()


def ensure_torch_available():
    if not TORCH_AVAILABLE:
        raise ImportError(
            f"This features requires a compatible torch version. {TORCH_IMPORT_ERROR}"
        )



import apache_beam as beam
from pathlib import Path
from apache_beam.io.fileio import MatchAll
from apache_beam.testing.test_pipeline import TestPipeline as BeamTestPipeline
from apache_beam.testing.util import assert_that, equal_to


def test_apache_beam_MatchAll():
    """Klay_beam version 0.12.x has a custom transform called MultiMatchFiles
    used to match multiple file patterns. However, this method is slow, because
    it could not parallelize.

    This test documents a better way to do this using Beam's MatchAll transform.

    See:
    https://beam.apache.org/releases/pydoc/current/apache_beam.io.fileio.html#apache_beam.io.fileio.MatchAll

    klay_beam.transforms.MultiMatchFiles should be deprecated after version
    0.12.x.
    """
    this_file = Path(__file__)
    data_dir = this_file.parent / "test_data"
    patterns = [str(data_dir / fn) for fn in ["00/*.txt", "01/*.txt"]]
    expected = [
        str(data_dir / fn) for fn in ["00/a.txt", "00/b.txt", "01/a.txt", "01/b.txt"]
    ]

    with BeamTestPipeline() as p:
        results = p | beam.Create(patterns) | MatchAll()

        # Extract file names from the metadata for assertion
        file_names = results | "Extract File Names" >> beam.Map(
            lambda metadata: metadata.path
        )

        # Check if the results match the expected file names
        assert_that(file_names, equal_to(expected))



import numpy as np
from klay_beam.torch_transforms import ResampleTorchaudioTensor

from .utils import skip_if_no_torch
from klay_beam.torch_utils import TORCH_AVAILABLE


if TORCH_AVAILABLE:
    import torch


def test_interleave():
    """Verify that we can use numpy's .T property to interleave stereo audio"""

    # This is how stereo audio is normally stored with the rightmost dimension
    # containing audio arrays ready for DSP.
    deinterleaved_audio = np.array(
        [[10, 20, 30, 40, 50, 60, 70, 80, 90], [-1, -2, -3, -4, -5, -6, -7, -8, -9]]
    )

    # However, when passing audio to pydub, it needs to be interleaved, that is,
    # stored like this
    expected_result = np.array(
        [
            [10, -1],
            [20, -2],
            [30, -3],
            [40, -4],
            [50, -5],
            [60, -6],
            [70, -7],
            [80, -8],
            [90, -9],
        ]
    )

    actual_result = deinterleaved_audio.T

    assert np.array_equal(
        actual_result, expected_result
    ), f"expected {actual_result} to equal {expected_result}"


def test_multichannel_interleave():
    """Verify that we can use numpy's .T property to interleave multi-channel audio"""

    deinterleaved_audio = np.array(
        [
            [10, 20, 30, 40, 50, 60, 70, 80, 90],
            [-1, -2, -3, -4, -5, -6, -7, -8, -9],
            [50, 50, 50, 50, 50, 50, 50, 50, 50],
        ]
    )

    expected_result = np.array(
        [
            [10, -1, 50],
            [20, -2, 50],
            [30, -3, 50],
            [40, -4, 50],
            [50, -5, 50],
            [60, -6, 50],
            [70, -7, 50],
            [80, -8, 50],
            [90, -9, 50],
        ]
    )

    actual_result = deinterleaved_audio.T

    assert np.array_equal(
        actual_result, expected_result
    ), f"expected {actual_result} to equal {expected_result}"


@skip_if_no_torch
def test_resample():
    """Verify that we can resample both numpy and torch audio (to a torch.Tensor))"""
    stereo_audio_tensor = ("key1", torch.randn(2, 1000), 1000)
    stereo_audio_numpy = ("key2", stereo_audio_tensor[1].numpy(), 1000)

    resampleDoFn = ResampleTorchaudioTensor(target_sr=10, source_sr_hint=1000)

    _, resampled_stereo_audio_tensor, _ = resampleDoFn.process(stereo_audio_tensor)[0]
    assert resampled_stereo_audio_tensor.shape == (
        2,
        10,
    ), "Failed to resample torch.tensor to sr=10"
    assert isinstance(
        resampled_stereo_audio_tensor, torch.Tensor
    ), "Failed to return a torch.Tensor"

    _, resampled_stereo_audio_numpy, _ = resampleDoFn.process(stereo_audio_numpy)[0]
    assert resampled_stereo_audio_numpy.shape == (
        2,
        10,
    ), "Failed to resample ndarray to sr=10"
    assert isinstance(
        resampled_stereo_audio_numpy, torch.Tensor
    ), "Failed to return a torch.tensor"


@skip_if_no_torch
def test_resample_to_numpy():
    """Verify that we can resample both numpy and torch audio (to a numpy.ndarray)"""
    stereo_audio_tensor = ("key1", torch.randn(2, 1000), 1000)
    stereo_audio_numpy = ("key2", stereo_audio_tensor[1].numpy(), 1000)

    resampleDoFn = ResampleTorchaudioTensor(target_sr=10, output_numpy=True)

    _, resampled_stereo_audio_tensor, _ = resampleDoFn.process(stereo_audio_tensor)[0]
    assert resampled_stereo_audio_tensor.shape == (
        2,
        10,
    ), "Failed to resample torch.tensor to sr=10"
    assert isinstance(
        resampled_stereo_audio_tensor, np.ndarray
    ), "Failed to return a torch.ndarray"

    _, resampled_stereo_audio_numpy, _ = resampleDoFn.process(stereo_audio_numpy)[0]
    assert resampled_stereo_audio_numpy.shape == (
        2,
        10,
    ), "Failed to resample ndarray to sr=10"
    assert isinstance(
        resampled_stereo_audio_numpy, np.ndarray
    ), "Failed to return a np.ndarray"



import argparse
import apache_beam as beam
from apache_beam.options.pipeline_options import PipelineOptions
from apache_beam.options.pipeline_options import SetupOptions
import logging
import torch
import sys
import os
import os.path


"""
Check if CUDA is available and log debug information.

Creating a docker image with CUDA support can be tricky. See the klay_beam
GitHub actions for an example of how to build such an image.

When running on Dataflow, the `--sdk_container_image` needs pytorch to be
installed. Unlike most klay_beam jobs, the image does not actually need the
klay_beam package installed because this file does not actually import klay_beam
(klay_beam should still be installed in the local launch environment).

Try using a local development environment such as:

conda env create -f environment/py3.10-torch1.11-cuda11.3.yml
conda env create -f environment/py3.10-torch1.11.yml
conda activate <your-env-name>

pip install apache_beam==2.51.0 klay_beam==0.12.2

# You can test your local environment
python -m klay_beam.run_cuda_test --runner Direct

# To run on Dataflow, configure the following environment variables as per the
# klay_beam README:
GCP_PROJECT_ID=<your project id>
GCP_SA_EMAIL=<your service account email>
TEMP_GS_URL=<gs:// path to a temporary directory>

python -m klay_beam.run_cuda_test \
    --region us-east4 \
    --autoscaling_algorithm NONE \
    --runner DataflowRunner \
    --service_account_email ${GCP_SA_EMAIL} \
    --disk_size_gb 50 \
    --experiments use_runner_v2 \
    --sdk_container_image \
        klaymusic/klay-beam:0.12.2-py3.10-beam2.51.0-torch1.11-cuda11.3 \
    --sdk_location container \
    --temp_location ${TEMP_GS_URL} \
    --project ${GCP_PROJECT_ID} \
    --dataflow_service_options \
        "worker_accelerator=type:nvidia-tesla-t4;count:1;install-nvidia-driver" \
    --job_name 'cuda-test'
"""


def log_debug_info(element):
    logging.info("python location: %s", sys.executable)
    logging.info("python version: %s", sys.version)
    logging.info("python path: %s", ":".join(sys.path))
    is_available = torch.cuda.is_available()
    logging.info("Checking CUDA availability!")
    logging.info("CUDA availability: %s", is_available)
    logging.info("CUDA device count: %s", torch.cuda.device_count())
    logging.info("pytorch was built with CUDA version: %s", torch.version.cuda)

    if is_available:
        logging.info("CUDA current device: %s", torch.cuda.current_device())

    try:
        # Create a dummy tensor
        tensor = torch.tensor([1.0])
        # Try to send the tensor to the GPU
        _ = tensor.cuda()
        logging.info("Successfully created a tensor and moved it to GPU.")
    except Exception as e:
        logging.error("Error during tensor creation or GPU transfer: %s", str(e))
    return element


def log_directory_contents(dir_path):
    try:
        # Check if the directory exists
        logging.info(f"-----------scanning directory: {dir_path}")
        if os.path.exists(dir_path):
            contents = os.listdir(dir_path)
            for fn in contents:
                logging.info(os.path.join(dir_path, fn))
    except Exception as e:
        # Catch any other errors (like permission errors)
        logging.info(f"Error while accessing {dir_path}: {e}")


def list_some_directories(element):
    log_directory_contents("/usr/local/nvidia")
    log_directory_contents("/usr/local/nvidia/lib64")
    log_directory_contents("/usr/local/cuda/lib64")
    log_directory_contents("/opt/apache/beam/")
    return element


def parse_args():
    parser = argparse.ArgumentParser()
    return parser.parse_known_args(None)


def run():
    known_args, pipeline_args = parse_args()
    print("known_args: {}".format(known_args))
    print("pipeline_args: {}".format(pipeline_args))

    # An example from the Apache Beam documentation uses the save_main_session
    # option. They describe the motivation for this option:
    #
    # > because one or more DoFn's in this workflow rely on global context
    # > (e.g., a module imported at module level).
    pipeline_options = PipelineOptions(pipeline_args)
    pipeline_options.view_as(SetupOptions).save_main_session = True

    with beam.Pipeline(options=pipeline_options) as p:
        (
            p
            | "CreateSingleton" >> beam.Create([None])
            | "CudaDebug" >> beam.Map(log_debug_info)
            | "ListDirectories" >> beam.Map(list_some_directories)
        )


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO)
    run()



from apache_beam.io.filesystem import FileMetadata
from pathlib import Path
from klay_beam.transforms import SkipCompleted


def test_skip_completed():
    this_file = Path(__file__)
    data_dir = str(this_file.parent / "test_data")

    skips = SkipCompleted(
        old_suffix=".wav",
        new_suffix=".01.txt",
    )
    metadata_01 = FileMetadata(
        path=str(Path(data_dir) / "exists.wav"),
        size_in_bytes=16,
    )
    assert (
        skips.process(metadata_01) == []
    ), "SkipCompleted should return and empty list when the target file is present"

    metadata_02 = FileMetadata(
        path="/dummy/dir/exists.wav",
        size_in_bytes=16,
    )
    skips = SkipCompleted(
        old_suffix=".wav",
        new_suffix=[".01.txt", ".02.txt", ".03.txt"],
        source_dir="/dummy/dir",
        target_dir=data_dir,
    )
    assert (
        skips.process(metadata_02) == []
    ), "SkipCompleted should return and empty list when all the target files are present"

    does_not_skip = SkipCompleted(
        old_suffix=".wav",
        new_suffix=[".01.txt", ".02.txt", ".03.txt", ".04.txt"],
        source_dir="/dummy/dir",
        target_dir=data_dir,
    )
    assert does_not_skip.process(metadata_02) == [
        metadata_02
    ], "SkipCompleted should return the input file when some target files are missing"

    # test overwrite
    metadata_03 = FileMetadata(
        path=str(Path(data_dir) / "exists.wav"),
        size_in_bytes=16,
    )
    does_not_skip = SkipCompleted(
        old_suffix=".wav",
        new_suffix=".01.txt",
        overwrite=True,
    )
    assert does_not_skip.process(metadata_03) == [
        metadata_03
    ], "SkipCompleted should return the input file when overwrite is True"


def test_skip_completed_with_timestamp():
    this_file = Path(__file__)
    data_dir = str(this_file.parent / "test_data")
    does_not_skip = SkipCompleted(
        old_suffix=".wav",
        new_suffix=".01.txt",
        check_timestamp=True,
    )
    source1 = FileMetadata(
        path=str(Path(data_dir) / "exists.wav"),
        size_in_bytes=16,
        last_updated_in_seconds=float("inf"),
    )
    assert does_not_skip.process(source1) == [source1], (
        "SkipCompleted with `check_timestamp=True`should return the input file when the target "
        "file is present but older than the source file"
    )

    source2 = FileMetadata(
        path=str(Path(data_dir) / "exists.wav"),
        size_in_bytes=16,
        last_updated_in_seconds=1,
    )
    assert does_not_skip.process(source2) == [], (
        "SkipCompleted with `check_timestamp=True`should return an empty list when the target "
        "file  is present and newer than the source file"
    )



import pathlib
import io
from typing import Optional, Type, Union, List, Tuple
from packaging import version as packaging_version
import apache_beam as beam
from apache_beam.io import filesystems
import apache_beam.io.fileio as beam_io
import pydub
import soundfile as sf
import numpy as np
import logging
import librosa

from apache_beam.io.filesystem import FileMetadata
from apache_beam.io.filesystems import FileSystems

from .path import move, remove_suffix


def numpy_to_pydub_audio_segment(
    audio_data: np.ndarray, sr: int, bit_depth=16
) -> pydub.AudioSegment:
    assert audio_data.ndim == 2, "audio_data must be 2 dimensional"
    channels, _ = audio_data.shape

    assert channels in [
        1,
        2,
    ], f"PyDub only supports mono and stereo audio (found {channels} channels)"

    # PuDub only supports 8 and 16 bit depth.
    assert bit_depth in [8, 16], f"bit_depth must be 8 or 16 (found {bit_depth}))"

    audio_data = audio_data.T  # pydub expects interleaved audio
    audio_data = audio_data * (2 ** (bit_depth - 1) - 1)  # scale to bit_depth.

    np_type: Optional[Type[np.signedinteger]] = None
    sample_width: Optional[int] = None
    if bit_depth == 8:
        np_type = np.int8
        sample_width = 1
    elif bit_depth == 16:
        np_type = np.int16
        sample_width = 2
    else:
        assert False, f"bit_depth must be 8, 16 (found {bit_depth}))"

    audio_data = audio_data.astype(np_type)
    raw_audio_buffer = io.BytesIO(audio_data.tobytes())

    audio_segment = pydub.AudioSegment.from_raw(
        raw_audio_buffer, frame_rate=sr, channels=channels, sample_width=sample_width
    )

    return audio_segment


def numpy_to_mp3(
    audio_data: np.ndarray,
    sr: int,
    bitrate="256k",
    intermediary_bit_depth=16,
):
    """Convert the audio data to an in-memory mp3 encoded file-like object."""
    channels, samples = audio_data.shape
    assert (
        samples > channels
    ), "incoming audio data must not be interleaved (last dimension must be audio)"

    logging.info(
        "Creating mp3: length: {:.3f} seconds. Channels: {}".format(
            samples / sr, channels
        )
    )

    # This needs to write a file-like object in memory. Loading file-like objects
    # is now supported by torchaudio (both sox_io and soundfile backends), but
    # torchaudio.save can only write to disk. As a result, we use pydub, which
    # invokes ffmpeg to write to memory.

    # For pytorch support see: https://pytorch.org/audio/stable/backend.html
    # For pydub docs see: https://github.com/jiaaro/pydub
    audio_segment = numpy_to_pydub_audio_segment(
        audio_data, sr, bit_depth=intermediary_bit_depth
    )

    # create an in-memory file-like object to write to
    in_memory_file_buffer = io.BytesIO()

    # Encode the pydub.AudioSegment as an mp3 (in-memory).
    #
    # ```
    # # Figure out which encoders are available (libmp3lame):
    # ffmpeg -encoders | grep mp3
    #
    # # Identify supported bit-rates
    # ffmpeg -h encoder=libmp3lame
    # ```
    audio_segment.export(in_memory_file_buffer, format="mp3", bitrate=bitrate)
    in_memory_file_buffer.seek(0)
    return in_memory_file_buffer


# `pip install soundfile` has buggy vorbis support. soundfile can only
# reliably write ogg files when the underlying libsndfile is 1.2.0 or greater
# You can check this with `soundfile.__libsndfile_version__`. For details,
# see: https://github.com/bastibe/python-soundfile/issues/130
sf_current_version = packaging_version.parse(sf.__libsndfile_version__)
sf_required_version = packaging_version.parse("1.2.0")
sf_version_ok = sf_current_version >= sf_required_version


def numpy_to_ogg(audio_data: np.ndarray, sr: int, safe=True):
    """Convert the audio data to an in-memory ogg encoded file-like object using
    the soundfile library.

    Due to limitations in the available libraries for writing ogg files,
    numpy_to_ogg requires libsndfile 1.2.0 or greater to work reliably. When
    invoked, this function will check the version of libsndfile and raise an
    error if an older version is found. You can disable this check by calling
    with `safe=False`. For details, see:
    https://github.com/bastibe/python-soundfile/issues/130
    """

    # Why use the soundfile package instead of pydub?
    # feature gaps in pydub, ffmpeg, pip's soundfile, and libsndfile.

    # We could use pydub, but need to ensure that the underlying ffmpeg command
    # is compiled with vorbis support. This is not the case when installing with
    # `conda install ffmpeg` on an Intel Mac in May 2023.

    assert audio_data.ndim == 2, "audio_data must be 2 dimensional"
    channels, samples = audio_data.shape
    logging.info(
        "Creating ogg: length: {:.3f} seconds. Channels: {}".format(
            samples / sr, channels
        )
    )

    if not sf_version_ok:
        error_message = (
            f"Old libsndfile. Found:{sf_current_version} need:{sf_required_version}"
        )
        if safe:
            assert False, error_message
        else:
            logging.warning(error_message)

    in_memory_file_buffer = io.BytesIO()
    sf.write(in_memory_file_buffer, audio_data.T, sr, subtype="VORBIS", format="OGG")
    in_memory_file_buffer.seek(0)
    return in_memory_file_buffer


def numpy_to_wav(audio_data: np.ndarray, sr: int, bit_depth=16):
    subtype = None
    if bit_depth == 8:
        subtype = "PCM_8"
    elif bit_depth == 16:
        subtype = "PCM_16"
    elif bit_depth == 24:
        subtype = "PCM_24"
    elif bit_depth == 32:
        subtype = "FLOAT"
    elif bit_depth == 64:
        subtype = "DOUBLE"
    else:
        raise ValueError(f"bit_depth must be 8, 16, 24, 32, or 64 (found {bit_depth}))")

    channels, samples = audio_data.shape
    logging.info(
        "Creating wav: length: {:.3f} seconds. Channels: {}".format(
            samples / sr, channels
        )
    )

    in_memory_file_buffer = io.BytesIO()
    sf.write(in_memory_file_buffer, audio_data.T, sr, format="WAV", subtype=subtype)
    in_memory_file_buffer.seek(0)
    return in_memory_file_buffer


def numpy_to_file(numpy_data: np.ndarray):
    in_memory_file_buffer = io.BytesIO()
    np.save(in_memory_file_buffer, numpy_data)
    in_memory_file_buffer.seek(0)
    return in_memory_file_buffer


def extract_wds_id_and_ext(readable_file: beam_io.ReadableFile):
    """Given an apache_beam ReadableFile, get a WebDataset style id and
    extension, where the id is everything up the first dot in the filename, and
    the extension is everything after the first dot in the filename.

    Can be used with beam.Map(extract_wds_id_and_ext)
    """
    path = pathlib.Path(readable_file.metadata.path)
    id = str(path.parent / path.name.split(".")[0])
    ext = ".".join(path.name.split(".")[1:])
    return (id, ext)


def write_file(output_path_and_buffer: Tuple[str, io.BytesIO]):
    """Helper function for writing a buffer to a given path. This should be able
    to handle gs:// style paths as well as local paths.

    Can be used with beam.Map(write_file)
    """
    output_path, buffer = output_path_and_buffer
    logging.info("Writing to: {}".format(output_path))
    with filesystems.FileSystems.create(output_path) as file_handle:
        file_handle.write(buffer.read())


class SkipCompleted(beam.DoFn):
    def __init__(
        self,
        old_suffix: str,
        new_suffix: Union[str, List[str]],
        source_dir: Optional[str] = None,
        target_dir: Optional[str] = None,
        check_timestamp: bool = False,
        overwrite: bool = False,
    ):
        if isinstance(new_suffix, str):
            new_suffix = [new_suffix]
        self._new_suffixes = new_suffix
        self._old_suffix = old_suffix

        assert (source_dir is None) == (
            target_dir is None
        ), "source_dir and target_dir must both be None or strings"

        self._source_dir = source_dir
        self._target_dir = target_dir
        self._check_timestamp = check_timestamp
        self._overwrite = overwrite

    def process(self, source_metadata: FileMetadata):  # type: ignore
        if self._overwrite:
            return [source_metadata]

        check = remove_suffix(source_metadata.path, self._old_suffix)
        if self._source_dir is not None:
            check = move(check, self._source_dir, self._target_dir)
        checks = [check + suffix for suffix in self._new_suffixes]
        limits = [1 for _ in checks]

        results = FileSystems.match(checks, limits=limits)
        assert len(results) > 0, "Unexpected empty results. This should never happen."

        for result in results:
            num_matches = len(result.metadata_list)
            logging.info(f"Found {num_matches} of: {result.pattern}")
            if num_matches != 0 and self._check_timestamp:
                for target_metadata in result.metadata_list:
                    if (
                        target_metadata.last_updated_in_seconds
                        < source_metadata.last_updated_in_seconds
                    ):
                        logging.info(
                            f"Do not skip! A target was found ({target_metadata.path}), but it is "
                            f"older than source file ({source_metadata.path})"
                        )
                        return [source_metadata]
            elif num_matches == 0:
                return [source_metadata]

        logging.info(f"Targets already exist. Skipping: {source_metadata.path}")
        return []


class MultiMatchFiles(beam.PTransform):
    """Like beam.io.fileio.MatchFiles, but takes a list of patterns and returns
    a single PCollection of FileMetadata objects.

    Deprecated in klay_beam v0.13.0. Instead, use this instead:

    ```
    pipeline | beam.Create([patterns]) | beam.io.fileio.MatchAll()
    ```
    """

    def __init__(self, patterns: list[str]):
        raise NotImplementedError(
            f"Matching {patterns} failed. "
            "MultiMatchFiles was removed in klay_beam v0.13.0. Use "
            "`beam.Create([patterns]) | beam.io.fileio.MatchAll()` instead."
        )


class LoadWithLibrosa(beam.DoFn):
    """Use librosa to load audio files to numpy arrays."""

    def __init__(self, target_sr: Optional[int], mono: bool):
        self.target_sr = target_sr
        self.mono = mono

    def process(self, readable_file: beam_io.ReadableFile):  # type: ignore
        """
        Given an Apache Beam ReadableFile, return a `(input_filename, a, sr)`
        tuple where:
            - `input_filename` is a string
            - `a` is a numpy array
            - `sr` is an int

        For a stereo audio file named '/path/to.some/file.key.mp3', return
        ```
        (
            '/path/to.some/file.key.mp3',
            np.array([[0.1, 0.2, 0.3], [0.1, 0.2, 0.3]]),
            44100
        )
        ```

        WARNING! Librosa is inconsistent about the shape of audio array. When
        mono=True, AND when the audio file is a mono file, librosa.load
        returns a 1-D numpy array. Librosa ONLY returns a 2-D array when the
        input audio file has multiple channels AND mono=False.
        """
        path = pathlib.Path(readable_file.metadata.path)

        # get the file extension without a period in a safe way
        ext_without_dot = path.suffix.lstrip(".")
        ext_without_dot = None if ext_without_dot == "" else ext_without_dot

        file_like = readable_file.open(mime_type="application/octet-stream")
        audio_array = None

        logging.info("Loading: {}".format(path))
        try:
            audio_array, sr = librosa.load(file_like, sr=self.target_sr, mono=self.mono)
            if self.target_sr is not None:
                assert sr == self.target_sr
        except RuntimeError as e:
            # We don't want to log the stacktrace, but for debugging, here's how
            # we could access it we can access it:
            #
            # import traceback
            # tb_str = traceback.format_exception(
            #     etype=type(e), value=e, tb=e.__traceback__
            # )
            logging.warning(f"Error loading audio: {path}")
            # return [beam.pvalue.TaggedOutput("failed", (str(path), e))]
            return [beam.pvalue.TaggedOutput("failed", (str(path), e))]

        num_samples = (
            len(audio_array) if audio_array.ndim == 1 else audio_array.shape[1]
        )

        logging.warning(
            f"Loaded {num_samples / sr:.3f} second, "
            f"{audio_array.shape}-shaped audio from: {path}"
        )

        return [(readable_file.metadata.path, audio_array, sr)]



import argparse
import os.path
import logging

import apache_beam as beam
import apache_beam.io.fileio as beam_io
from apache_beam.options.pipeline_options import PipelineOptions
from apache_beam.options.pipeline_options import SetupOptions

from .torch_transforms import LoadWithTorchaudio


"""
# Example usage:

```bash
python -m klay_beam.run_example \
    --runner Direct \
    --source_audio_suffix .mp3 \
    --source_audio_path \
        '/local/path/to/mp3s/' \

python -m klay_beam.run_example \
    --runner Direct \
    --source_audio_suffix .mp3 \
    --source_audio_path ${AUDIO_URL}


KLAY_BEAM_CONTAINER=us-docker.pkg.dev/<your-gcp-project>/<your-docker-artifact-registry>/<your-docker-image>:<tag>
SERVICE_ACCOUNT_EMAIL=dataset-dataflow-worker@klay-training.iam.gserviceaccount.com
TEMP_GS_URL=gs://<your-gs-bucket>/<your-writable-dir/>
AUDIO_URL='gs://<your-audio-bucket>/audio/'


python -m klay_beam.run_example \
    --runner DataflowRunner \
    --max_num_workers=128 \
    --region us-central1 \
    --autoscaling_algorithm THROUGHPUT_BASED \
    --service_account_email ${SERVICE_ACCOUNT_EMAIL} \
    --experiments=use_runner_v2 \
    --sdk_container_image ${KLAY_BEAM_CONTAINER} \
    --sdk_location=container \
    --temp_location ${TEMP_GS_URL} \
    --project klay-training \
    --source_audio_suffix .mp3 \
    --source_audio_path ${AUDIO_URL} \
    --machine_type n1-standard-8 \
    --job_name 'example-job-000'
```
"""


def parse_args():
    parser = argparse.ArgumentParser()

    parser.add_argument(
        "--source_audio_path",
        dest="input",
        required=True,
        help="""
        Specify the parent audio file directory. This can be a local path or a gs:// URI.

        To get only some wav files, try:
        '/Users/alice/datasets/fma_large/005/'

        To run on the full dataset use:
        'gs://klay-datasets/mtg-jamendo/'
        """,
    )

    parser.add_argument(
        "--source_audio_suffix",
        required=True,
        choices=[".mp3", ".wav", ".aif", ".aiff"],
        help="""
        Which format are candidate audio files saved with?
        """,
    )

    return parser.parse_known_args(None)


def run():
    known_args, pipeline_args = parse_args()
    print("known_args: {}".format(known_args))
    print("pipeline_args: {}".format(pipeline_args))

    # pickle the main session in case there are global objects
    pipeline_options = PipelineOptions(pipeline_args)
    pipeline_options.view_as(SetupOptions).save_main_session = True

    # Pattern to recursively find mp3s inside source_audio_path
    match_pattern = os.path.join(
        known_args.input, f"**{known_args.source_audio_suffix}"
    )

    with beam.Pipeline(argv=pipeline_args, options=pipeline_options) as p:
        _, failed, durations = (
            p
            # MatchFiles produces a PCollection of FileMetadata objects
            | beam_io.MatchFiles(match_pattern)
            # Prevent "fusion". See:
            # https://cloud.google.com/dataflow/docs/pipeline-lifecycle#preventing_fusion
            | "Reshuffle" >> beam.Reshuffle()
            # ReadMatches produces a PCollection of ReadableFile objects
            | beam_io.ReadMatches()
            | "Load Audio"
            >> beam.ParDo(LoadWithTorchaudio()).with_outputs(
                "failed", "duration_seconds", main="audio"
            )
        )

        (
            durations
            | "SumLengths" >> beam.CombineGlobally(sum)
            | "LogDuration"
            >> beam.Map(
                lambda x: logging.info(
                    "Total duration of loaded audio: "
                    f"~= {x:.3f} seconds "
                    f"~= {x / 60:.3f} minutes "
                    f"~= {x / 60 / 60:.3f} hours"
                )
            )
        )

        (
            failed
            | "Log Failed" >> beam.Map(lambda x: logging.warning(x))
            | "Count" >> beam.combiners.Count.Globally()
            | "Log Failed Count"
            >> beam.Map(lambda x: logging.warning(f"Failed to decode {x} files"))
        )


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO)
    run()



import torch
from pkg_resources import parse_version


def get_device(mps_valid: bool = False) -> torch.device:
    # torch.has_mps is not available in older versions of torch such as 1.11.
    # Ensure it exists before using it. This also satisfies mypy.
    if mps_valid and hasattr(torch, "has_mps") and torch.has_mps:
        return torch.device("mps")
    elif torch.cuda.is_available():
        # if cuda version < 11, RTX 30xx series not available
        if parse_version(torch.version.cuda) < parse_version("11"):  # type: ignore
            for i in range(torch.cuda.device_count()):
                if "RTX 30" not in torch.cuda.get_device_name(i):
                    return torch.device("cuda", i)
        else:
            return torch.device("cuda")
    return torch.device("cpu")



import setuptools

# DO NOT DELETE THIS FILE!
#
# Even though it is just a stub, it is required to correctly package the
# klay_beam module for running on Dataflow. We use the dataflow --setup_file
# option to specify this file when launching a Dataflow job. This is how we
# instruct (locally running) python code to create an sdist package and upload
# it to GCS. The sdist package is then downloaded by the workers.

if __name__ == "__main__":
    setuptools.setup()



import pytest
import librosa
import numpy as np
import scipy

from klay_beam.transforms import (
    numpy_to_mp3,
    numpy_to_ogg,
    numpy_to_wav,
)


def create_test_sine(
    hz: float = 1000,
    sr: int = 44100,
    total_seconds: float = 0.5,
    fade_in_seconds: float = 0,
):
    """Create a 0.5 second mono audio signal with a sine wave

    Note the dimensionality. This is just a single channel in a single dimension
    (time) i.e. `[0, 0, 0, ...]`. Some functions in `klay_beam` may expect audio
    to be in a multi-channel format, (i.e. mono like this: `[[0, 0, 0, ...]]`).
    """

    total_samples = int(sr * total_seconds)
    fade_in_samples = int(sr * fade_in_seconds)
    fade_in_samples = min(total_samples, fade_in_samples)

    t = np.linspace(0, total_seconds, total_samples, endpoint=False)
    audio_channel = np.sin(2 * np.pi * hz * t)

    fade_in = np.linspace(0.0, 1.0, fade_in_samples) ** 2.0
    audio_channel[:fade_in_samples] *= fade_in

    return audio_channel, sr


def total_difference(a, b):
    # calculate the difference between two numpy arrays, returning a scalar, no
    # matter what the input rank is
    assert a.shape == b.shape
    return np.sum(np.abs(a - b))


def max_difference(a, b):
    assert a.shape == b.shape
    return np.max(np.abs(a - b))


def average_difference(a, b):
    assert a.shape == b.shape
    return np.average(np.abs(a - b).reshape(-1))


def spectral_divergence(a, b):
    """Compare the spectral magnitude of two signals. This effectively compares
    two signals in the frequency domain. The exact formula is from
    [Arik et al., 2018](https://arxiv.org/abs/1808.06719), but I've seen similar
    calculations used elsewhere.

    Why bother with a spectral comparison? Heuristic data compressors such as
    mp3 will make significant changes to the phase (and content) of a signal,
    but dominant frequency bands should be mostly preserved. We can use this to
    identify if (for example) our file writer replaced music with noise or
    silence. It is simple check that should also be robust against the small
    time shifts introduced by mp3 encoding.

    Empirically, two audio files with the same musical content, but
    different file encodings (e.g. wav vs mp3) will have a divergence of < 0.1.
    some examples:

    ```
    x_wav, _ = librosa.load("tests/test_data/music/01.wav", mono=False)
    x_mp3, _ = librosa.load("tests/test_data/music/01.mp3", mono=False)
    x_ogg, _ = librosa.load("tests/test_data/music/01.ogg", mono=False)
    x_opus, _ = librosa.load("tests/test_data/music/01.opus", mono=False)
    y_mp3, _ = librosa.load("tests/test_data/music/02.mp3", mono=False)

    # mp3 vs wav
    spectral_divergence(x_wav, x_mp3)
    [0.0395, 0.0385]

    # ogg vs wav
    spectral_divergence(x_wav, x_ogg)
    [0.0710, 0.0753]

    # different audio content
    spectral_divergence(x_wav, y_wav)
    [1.0358, 1.0477]
    ```

    However, it should be noted that these values will be affected by
    bitrate, the exact encoder used, and other parameters.

    Args:
        a, b: Two numpy arrays containing time-domain audio. Each array should
        have shape=(channels, samples). Note that this operation is not
        commutative (`spectral_divergence(a, b) != spectral_divergence(b, a)`).

    Returns:
        divergences: A numpy array of spectral divergence values, with one value
        for each channel. 0 means the signals are identical, 1 means they are
        quite different. The value represents an average across the entire
        signal (longer files will not have a higher value than shorter files).
    """
    assert a.shape == b.shape
    assert a.ndim == 2

    # apply an fft to each channel of each signal
    a_magnitudes = [np.abs(scipy.signal.stft(channel)[2]) for channel in a]
    b_magnitudes = [np.abs(scipy.signal.stft(channel)[2]) for channel in b]

    return np.array(
        [
            np.linalg.norm(a - b) / np.linalg.norm(a)
            for a, b in zip(a_magnitudes, b_magnitudes)
        ]
    )


def test_sine():
    sine, _ = create_test_sine(hz=1, sr=4, total_seconds=1)
    expected_sine = np.array([0, 1, 0, -1])

    assert sine.shape == (4,)
    assert total_difference(sine, expected_sine) < 1e-14


def test_total_difference():
    sine_a, _ = create_test_sine()
    sine_b, _ = create_test_sine()

    # Make sine_b different from sine_a by a total of 1.0
    sine_b[0] -= 0.5
    sine_b[1] += 0.5

    assert total_difference(sine_a, sine_a) == 0
    assert total_difference(sine_a, sine_b) == 1.0


def test_max_difference():
    sine_a, _ = create_test_sine()
    sine_b, _ = create_test_sine()

    # Make sine_b different from sine_a by a total of 1.0
    sine_b[0] += 0.3
    sine_b[1] -= 0.75

    assert max_difference(sine_a, sine_a) == 0
    assert max_difference(sine_a, sine_b) == pytest.approx(0.75, 1e-14)
    sine_b[2] += 0.8
    assert max_difference(sine_a, sine_b) == pytest.approx(0.8, 1e-14)


def test_average_difference():
    sine_a, _ = create_test_sine()
    sine_b, _ = create_test_sine()

    sine_b -= 0.3

    assert average_difference(sine_a, sine_a) == 0
    assert average_difference(sine_a, sine_b) == pytest.approx(0.3, 1e-14)

    # Check that we are getting the average of the absolute differences
    sine_b, _ = create_test_sine()
    sine_b[: len(sine_b) // 2] += 0.3
    sine_b[len(sine_b) // 2 :] -= 0.3
    assert average_difference(sine_a, sine_b) == pytest.approx(0.3, 1e-14)

    # check it also works with stereo
    sine_a = np.array([sine_a, sine_a])
    sine_b = np.array([sine_b, sine_b])
    assert average_difference(sine_a, sine_a) == 0
    assert average_difference(sine_a, sine_b) == pytest.approx(0.3, 1e-14)


def test_spectral_divergence():
    sine_1k0, _ = create_test_sine(total_seconds=10, hz=1000)
    sine_1k1, _ = create_test_sine(total_seconds=10, hz=1100)
    sine_10k, _ = create_test_sine(total_seconds=10, hz=10000)

    sine_1k_stereo = np.stack([sine_1k0, sine_1k0])
    sine_1k1_stereo = np.stack([sine_1k1, sine_1k1])
    sine_10k_stereo = np.stack([sine_10k, sine_10k])

    noise_stereo = np.random.normal(size=sine_1k_stereo.shape)
    silence_stereo = np.zeros_like(sine_1k_stereo)

    # should be at least somewhat similar
    assert np.all(spectral_divergence(sine_1k_stereo, sine_1k_stereo) == 0)
    assert np.all(spectral_divergence(sine_1k_stereo, sine_1k1_stereo) < 0.5)

    # should be very different
    assert np.all(spectral_divergence(sine_1k_stereo, sine_10k_stereo) > 1.0)
    assert np.all(spectral_divergence(sine_1k_stereo, noise_stereo) > 1.0)
    assert np.all(spectral_divergence(sine_1k_stereo, silence_stereo) >= 1.0)


def test_wav_file():
    sine, sr = create_test_sine()
    sine *= 0.5

    sine_mono = np.array([sine])
    sine_stereo = np.array([sine, sine])

    # make in-memory audio files
    wav_mono = numpy_to_wav(sine_mono, sr)
    wav_stereo = numpy_to_wav(sine_stereo, sr)

    # use Librosa to read the in-memory audio files. Annoyingly, librosa returns
    # mono audio as a 1-d array. As a result, we must compare to sine instead of
    # sine_mono
    librosa_mono, sr1 = librosa.load(wav_mono, sr=None, mono=False)
    assert max_difference(sine, librosa_mono) < 1e-4
    assert average_difference(sine, librosa_mono) < 1e-4

    # Do the same check, but verify that it works with 24-bit audio. Note that
    # we expect a higher precision from 24-bit audio.
    wav_mono = numpy_to_wav(sine_mono, sr, bit_depth=24)
    librosa_mono, sr1 = librosa.load(wav_mono, sr=None, mono=False)
    assert max_difference(sine, librosa_mono) < 1e-6
    assert average_difference(sine, librosa_mono) < 1e-6

    # Try a stereo file
    librosa_stereo, sr2 = librosa.load(wav_stereo, sr=None, mono=False)
    assert max_difference(sine_stereo, librosa_stereo) < 1e-4
    assert average_difference(sine_stereo, librosa_stereo) < 1e-4

    # Double check that we are getting the same sample rate as we started with
    assert sr1 == sr
    assert sr2 == sr


def test_mp3_file():
    sine, sr = create_test_sine()
    sine *= 0.5

    sine_mono = np.array([sine])
    sine_stereo = np.array([sine, sine])

    # make in-memory audio files
    mp3_mono = numpy_to_mp3(sine_mono, sr)
    mp3_stereo = numpy_to_mp3(sine_stereo, sr)

    # use Librosa to read the in-memory audio files. Annoyingly, librosa returns
    # mono audio as a 1-d array. As a result, we must compare to sine instead of
    # sine_mono
    librosa_mono, sr1 = librosa.load(mp3_mono, sr=None, mono=False)
    assert max_difference(sine, librosa_mono) < 1e-2
    assert average_difference(sine, librosa_mono) < 1e-4

    # Also verify the intermediary_bit_depth flag works as expected.
    mp3_mono = numpy_to_mp3(sine_mono, sr, intermediary_bit_depth=8)
    librosa_mono, sr1 = librosa.load(mp3_mono, sr=None, mono=False)
    assert max_difference(sine, librosa_mono) < 1e-1
    assert average_difference(sine, librosa_mono) < 1e-2

    # Try a stereo file
    librosa_stereo, sr2 = librosa.load(mp3_stereo, sr=None, mono=False)
    assert max_difference(sine_stereo, librosa_stereo) < 1e-2
    assert average_difference(sine_stereo, librosa_stereo) < 1e-4

    assert np.all(spectral_divergence(librosa_stereo, sine_stereo) < 0.1)

    # Double check that we are getting the same sample rate as we started with
    assert sr1 == sr
    assert sr2 == sr


def test_ogg_file():
    x, sr = librosa.load("tests/test_data/music/01.wav", sr=None, mono=False)
    in_memory_ogg = numpy_to_ogg(x, int(sr), safe=False)
    librosa_stereo, _ = librosa.load(in_memory_ogg, sr=None, mono=False)
    assert np.all(spectral_divergence(x, librosa_stereo) < 0.125)
